# #--------------------------------------------------------------------------------
# # 참조 모듈 목록.
# #--------------------------------------------------------------------------------
# from __future__ import annotations
# from typing import Awaitable, Callable, Final, Generic, ItemsView, Iterable, Iterator, KeysView, Optional, Sequence, Type, TypeVar, Tuple, Union, ValuesView
# from typing import Any, List, Dict, Set
# from typing import cast, overload
# import builtins
# import os
# # import cx_Freeze # type: ignore
# # from cx_Freeze import Executable # type: ignore
# import win32com.client # type: ignore


# #--------------------------------------------------------------------------------
# # 전역 상수 목록.
# #--------------------------------------------------------------------------------
# NAME: str = "altava-3dsmax2023-plugin"
# VERSION: str = "0.0.1"
# CREATOR: str = "altavagroup"
# DESCRIPTION: str = "altava conversion 3dsmax2023 plugin"
# BDIST_MSI: str = "bdist_msi"
# BUILD_EXE: str = "build_exe"
# CURRENTFILEPATH: str = os.path.abspath(__file__)
# ROOTPATH: str = os.path.dirname(CURRENTFILEPATH)


# #--------------------------------------------------------------------------------
# # 파일 진입점.
# #--------------------------------------------------------------------------------
# if __name__ == "__main__":
# 	msi_path = "example.msi"  # 수정할 MSI 파일 경로
# 	file_id = "UpdateBatFile"  # File 테이블의 File ID
# 	component_id = "MainComponent"  # Component 테이블의 Component ID
# 	directory_id = "TARGETDIR"  # Component 테이블의 Directory ID
# 	sequence = 1  # 설치 순서	
# 	msiFilePath: str = os.path.join(ROOTPATH, msi_path)

# 	installer = win32com.client.Dispatch("WindowsInstaller.Installer")
# 	database = installer.OpenDatabase(msiFilePath, 3)

# 	tables_query = "SELECT * FROM `_Tables` WHERE Name = 'Component'"
# 	view = database.OpenView(tables_query)
# 	view.Execute(None)
# 	record = view.Fetch()
# 	if record:
# 		print("Component 테이블이 존재합니다.")
# 	else:
# 		print("Component 테이블이 없습니다.")

# 	tables_query = "SELECT * FROM `_Tables` WHERE Name = 'File'"
# 	view = database.OpenView(tables_query)
# 	view.Execute(None)
# 	record = view.Fetch()
# 	if record:
# 		print("File 테이블이 존재합니다.")
# 	else:
# 		print("File 테이블이 없습니다.")
# 		create_table_query = """
# 			CREATE TABLE File (
# 				File TEXT PRIMARY KEY,
# 				Component_ TEXT NOT NULL,
# 				FileName TEXT NOT NULL,
# 				FileSize INTEGER,
# 				Version TEXT,
# 				Language TEXT,
# 				Attributes INTEGER,
# 				Sequence INTEGER
# 			)
# 		"""
# 		view = database.OpenView(create_table_query)
# 		view.Execute(None)
# 		view.Close()

# 	try:
# 		sourceFilePath: str = os.path.join(ROOTPATH, "res", "msi", "update.bat")
# 		sourceFileSize = os.path.getsize(sourceFilePath)
# 		view = database.OpenView("INSERT INTO File (File, Component_, FileName, FileSize, Version, Language, Attributes, Sequence) VALUES (?, ?, ?, ?, ?, ?, ?, ?)")
# 		record = installer.CreateRecord(8)
# 		record.SetString(1, file_id)  # File ID
# 		record.SetString(2, component_id)  # Component ID
# 		record.SetString(3, "update.bat")  # File Name
# 		record.SetInteger(4, sourceFileSize)  # File Size
# 		record.SetString(5, None)  # Version
# 		record.SetString(6, None)  # Language
# 		record.SetInteger(7, 0)  # Attributes
# 		record.SetInteger(8, sequence)  # Sequence	
# 		view.Execute(record)
# 		view.Close()
# 		view = database.OpenView("INSERT INTO Component (Component, ComponentId, Directory_, Attributes, Condition, KeyPath) VALUES (?, ?, ?, ?, ?, ?)")
# 		record = installer.CreateRecord(6)
# 		record.SetString(1, component_id)  # Component ID
# 		record.SetString(2, "{12345678-1234-1234-1234-123456789ABC}")  # Component GUID
# 		record.SetString(3, directory_id)  # Directory ID
# 		record.SetInteger(4, 0)  # Attributes
# 		record.SetString(5, None)  # Condition
# 		record.SetString(6, file_id)  # KeyPath
# 		view.Execute(record)	
# 		view.Close()

# 		with open(sourceFilePath, "rb") as file:
# 			bytes = file.read()

# 		database.Streams(file_id).Write(bytes)
# 		database.Commit()
# 	except Exception as exception:
# 		builtins.print(exception)
#	# finally:
#	# 	database.Close()


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterable, Iterator, Optional, Sequence, Type, TypeVar, Tuple, Union
from typing import ItemsView, KeysView, ValuesView
from typing import Any, List, Dict, Set
from typing import cast, overload
import builtins
from enum import Enum


#--------------------------------------------------------------------------------
# Microsoft Installer Table Field Type.
#--------------------------------------------------------------------------------
class TableFieldType(Enum):
	TEXT = "TEXT"		# 가변 길이의 문자열. (MSI 내부에서 필드마다 각각 다른 고정값 존재)
	INTEGER = "INTEGER"	# 32비트 정수 값.
	LONG = "LONG"		# 64비트 정수 값.
	OBJECT = "OBJECT"	# 바이너리데이터. (BLOB)


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterable, Iterator, Optional, Sequence, Type, TypeVar, Tuple, Union
from typing import ItemsView, KeysView, ValuesView
from typing import Any, List, Dict, Set
from typing import cast, overload
import builtins
from .core import ITable, IDatabase, IQuery, IRecord, Installer


#--------------------------------------------------------------------------------
# 테이블.
#--------------------------------------------------------------------------------
class Table(ITable):
	#--------------------------------------------------------------------------------
	# 컬럼 정보 반환.
	#--------------------------------------------------------------------------------
	def GetColumns(self) -> list:
		tableName: str = self.GetTableName()
		database: IDatabase = self.GetDatabase()
		try:
			query: IQuery = Installer.CreateQuery(database, f"SELECT Name FROM _Columns WHERE Table = {tableName}")
			query.Execute()
		except Exception as exception:
			raise

		record: IRecord = query.Fetch()
		columns = list()
		while record:
			fieldName: str = record.GetFieldValueAsString(0)
			fieldTypeCode: int = record.GetFieldValueAsInteger(1)
			fieldType: str = "Other"
			if fieldTypeCode == 0:
				fieldType = "String"
			elif fieldTypeCode == 1:
				fieldType = "Integer"
			columns.append((fieldName, fieldType))
			record = query.Fetch()
		query.Close()
		return columns


	#--------------------------------------------------------------------------------
	# 모든 레코드 반환.
	#--------------------------------------------------------------------------------
	def GetAllRecords(self) -> list:
		tableName: str = self.GetTableName()
		database: IDatabase = self.GetDatabase()
		records = list()
		query: IQuery = Installer.CreateQuery(database, f"SELECT * FROM {tableName}")
		query.Execute(None)
		data = query.Fetch()
		while data:
			records.append(data)
			data = query.Fetch()
		query.Close()
		return records

	#--------------------------------------------------------------------------------
	# 레코드 추가.
	#--------------------------------------------------------------------------------
	def Insert(self, record: IRecord) -> None:
		tableName: str = self.GetTableName()
		database: IDatabase = self.GetDatabase()
		columns: list = self.GetColumns()
		columnFieldsString: str = str()
		inputFieldsString: str = str()
		count: int = len(columns)

		for index in range(count):
			column = columns[index]
			if index + 1 < count:
				columnFieldsString += f"{column}, "
				inputFieldsString += f"?, "
			else:
				columnFieldsString += column
				inputFieldsString += f"?"

		query: IQuery = Installer.CreateQuery(database, f"INSERT INTO {self.__tableName} ({columnFieldsString}) VALUES ({inputFieldsString})")
		query.Execute(record)
		query.Close()



	#--------------------------------------------------------------------------------
	# 레코드 제거.
	#--------------------------------------------------------------------------------
	def Remove(self, record: IRecord) -> None:
		pass


	#--------------------------------------------------------------------------------
	# 레코드 갱신.
	#--------------------------------------------------------------------------------
	def Update(self, record: IRecord) -> None:
		pass


	#--------------------------------------------------------------------------------
	# 레코드 업데이트.
	#--------------------------------------------------------------------------------
	def UpdateRecords(self, whereFieldName: str, whereFieldValue: Union[int, str, None], updateFieldName: str, updateFieldValue: Union[int, str, None])  -> None:
		tableName: str = self.GetTableName()
		database: IDatabase = self.GetDatabase()
		record: IRecord = Installer.CreateRecordFromFieldValues((updateFieldValue, whereFieldValue))
		view = Installer.CreateQuery(database, f"UPDATE {tableName} SET {updateFieldName} = ? WHERE {whereFieldName} = ?")
		view.Execute(record)
		view.Close()


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, ItemsView, Iterable, Iterator, KeysView, Optional, Sequence, Type, TypeVar, Tuple, Union, ValuesView
from typing import Any, List, Dict, Set
from typing import cast, overload
import builtins
import os
from msi import IDatabase, IQuery, IRecord, ITable, Installer, DatabaseScope, QueryScope, CustomActionType


#--------------------------------------------------------------------------------
# BuildProcess.
#--------------------------------------------------------------------------------
class BuildProcess:
	#--------------------------------------------------------------------------------
	# 출력.
	#--------------------------------------------------------------------------------
	@staticmethod
	def PrintTable(database: IDatabase, tableName: str) -> None:
		# 파일 테이블 출력.
		table: ITable = database.GetTable(tableName)
		for record in table.GetAllRecords():
			record: IRecord = cast(IRecord, record)
			fieldValues = Installer.CreateFieldValuesFromRecord(record)
			builtins.print(f"[{tableName}] {fieldValues}")


	#--------------------------------------------------------------------------------
	# 생성.
	#--------------------------------------------------------------------------------
	@staticmethod
	def CreateMSIToFile(msiFilePath: str, applicationGUID: str) -> None:
		# 생성.
		database: IDatabase = Installer.CreateDatabase(msiFilePath)

		# 프로퍼티 테이블.
		if database.ContainsTable("Property"):
			propertyTable: ITable = database.GetTable("Property")
			propertyTable.UpdateRecords("Property", "ProductName", "Value", "ALTAVA Group 3ds Max 2023 Plugin")
			propertyTable.UpdateRecords("Property", "ProductCode", "Value", applicationGUID)
			propertyTable.UpdateRecords("Property", "ProductVersion", "Value", "0.0.10")
			propertyTable.UpdateRecords("Property", "Manufacturer", "Value", "ALTAVA Group")
			propertyTable.UpdateRecords("Property", "ProductLanguage", "Value", 1033)
		else:
			with QueryScope(database, "CREATE TABLE Property (Property TEXT NOT NULL PRIMARY KEY, Value TEXT)") as query:
				query.Execute()
			with QueryScope(database, "INSERT INTO Property (Property, Value) VALUES (?, ?)") as query:
				query.Execute(("ProductName", "ALTAVA Group 3ds Max 2023 Plugin"))
				query.Execute(("ProductCode", applicationGUID))
				query.Execute(("ProductVersion", "0.0.10"))
				query.Execute(("Manufacturer", "ALTAVA Group"))
				query.Execute(("ProductLanguage", 1033))
		BuildProcess.PrintTable(database, "Property")

		# 디렉토리 테이블.
		with QueryScope(database, "INSERT INTO Directory (Directory, Directory_Parent, DefaultDir) VALUES (?, ?, ?)") as query:
			query.Execute(("TARGETDIR", None, "SourceDir"))
			query.Execute(("PROGRAMFILES", "TARGETDIR", ".:Programe Files"))
			query.Execute(("AUTODESK_3DSMAX2023", "PROGRAMFILES", "Autodesk\\3ds Max 2023"))
			query.Execute(("AUTODESK_3DSMAX2023_BIN_ASSEMBLIES", "AUTODESK_3DSMAX2023", "bin\\assemblies"))
			query.Execute(("AUTODESK_3DSMAX2023_SCRIPTS_ALTAVAMAXPLUGIN", "AUTODESK_3DSMAX2023", "scirpts\\AltavaMaxPlugin"))
			query.Execute(("AUTODESK_3DSMAX2023_SCRIPTS_STARTUP", "AUTODESK_3DSMAX2023", "scirpts\\Startup"))
		BuildProcess.PrintTable(database, "Directory")

		# 컴포넌트 테이블.
		with QueryScope(database, "INSERT INTO Component (Component, ComponentId, Directory_, Attributes, Condition, KeyPath) VALUES (?, ?, ?, ?, ?, ?)") as query:
			query.Execute(("ALTAVA_MAX_PLUGIN", Installer.CreateGUID(), "AUTODESK_3DSMAX2023_SCRIPTS_ALTAVAMAXPLUGIN", 0, "VersionNT >= 1000", "AUTODESK_3DSMAX2023_SCRIPTS_ALTAVAMAXPLUGIN"))
			query.Execute(("BABYLON_JS_EXPORTER", Installer.CreateGUID(), "AUTODESK_3DSMAX2023", 0, "VersionNT >= 1000", None))

		# 파일 테이블.
		database.AddStream
		with QueryScope(database, "INSERT INTO File (File, Component_, FileName, FileSize, Version, Language, Attributes, Sequence) VALUES (?, ?, ?, ?, ?, ?, ?, ?)") as query:
			fileSize = os.path.getsize("update.bat")
			query.Execute(("update.bat", "ALTAVA_MAX_PLUGIN", "update.bat", fileSize, None, None, 0, 1))
			# query.Execute(("update.bat", "BABYLON_JS_EXPORTER", "update.bat", fileSize, None, None, 0, 1))
		BuildProcess.PrintTable(database, "File")

		# 커스텀 액션 테이블.
		with QueryScope(database, "INSERT INTO CustomAction (Action, Type, Source, Target) VALUES (?, ?, ?, ?)") as query:
			query.Execute(("UpdateBatchFile", CustomActionType.EXECUTE_FILE_ID_BASED.value, "update.bat", ""))

		# # 설치 실행 시퀀스 테이블.
		BuildProcess.PrintTable(database, "InstallExecuteSequence")
		# with QueryScope(database, "INSERT INTO InstallExecuteSequence (Action, Condition, Sequence) VALUES (?, ?, ?)") as query:
		# 	query.Execute(("UpdateBatchFile", "NOT Installed", 700))

		# 최종 반영.
		database.Commit()
		database.Close()


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterable, Iterator, Optional, Sequence, Type, TypeVar, Tuple, Union
from typing import ItemsView, KeysView, ValuesView
from typing import Any, List, Dict, Set
from typing import cast, overload
import builtins
import os
import setuptools


#--------------------------------------------------------------------------------
# 전역 상수 목록.
#--------------------------------------------------------------------------------
UTF8: str = "utf-8"
FILE_READTEXT: str = "rt"


#--------------------------------------------------------------------------------
# 참조 메타 데이터 목록.
#--------------------------------------------------------------------------------
NAME: str = "xid-msi"
AUTHOR: str = "xidware"
AUTHOR_EMAIL: str = "developer@xidware.com"
DESCRIPTION: str = "Xidware Microsoft Installer"
LONG_DESCRIPTION_CONTENT_TYPE: str = "text/markdown"
URL: str = "https://xidware.com"
PYTHON_REQUIRES: str = ">=3.9.7" # 실제로는 3.10.11 이상.
LONGDESCRIPTION: str = str()
with open(file = "README.md", mode = FILE_READTEXT, encoding = UTF8) as file: LONGDESCRIPTION = file.read()
with open(file = "VERSION", mode = FILE_READTEXT, encoding = UTF8) as file: VERSION = file.read()


#--------------------------------------------------------------------------------
# 빌드.
#--------------------------------------------------------------------------------
setuptools.setup(
	name = NAME,
	version = VERSION,
	author = AUTHOR,
	author_email = AUTHOR_EMAIL,
	description = DESCRIPTION,
	long_description = LONGDESCRIPTION,
	long_description_content_type = LONG_DESCRIPTION_CONTENT_TYPE,
	url = URL,
	packages = setuptools.find_packages(where = "src"),
	include_package_data = True,
	package_dir = { "": "src" },
	package_data = {
		"": [
			"res/*"
		],
	},
	scripts = [

	],
	entry_points = {
		"console_scripts": [
			"msi=msi.commandmanager:CommandManager.Run"
		]
	},
	install_requires = [
		"xid-xpl"
	],
	classifiers = [
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent"
	],
	python_requires = PYTHON_REQUIRES
)


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterable, Iterator, Optional, Sequence, Type, TypeVar, Tuple, Union
from typing import ItemsView, KeysView, ValuesView
from typing import Any, List, Dict, Set
from typing import cast, overload
import builtins
from .core import IQuery, IRecord, Installer


#--------------------------------------------------------------------------------
# Microsoft Installer Database의 SQL 처리자. (View)
#--------------------------------------------------------------------------------
class Query(IQuery):
	#--------------------------------------------------------------------------------
	# 실행.
	#--------------------------------------------------------------------------------
	def Execute(self, value: Union[IRecord, Tuple, None] = None) -> None:
		try:
			if value == None:
				rawdata = self.GetRawData()
				builtins.print(f"{rawdata}.Execute(None)")
				rawdata.Execute(None)
			elif builtins.isinstance(value, IRecord):
				rawdata = self.GetRawData()
				builtins.print(f"{rawdata}.Execute({value})")
				rawdata.Execute(value.GetRawData())
			elif builtins.isinstance(value, Tuple):
				record: IRecord = Installer.CreateRecordFromFieldValues(value)
				rawdata = self.GetRawData()
				builtins.print(f"{rawdata}.Execute({record})")
				rawdata.Execute(record.GetRawData())
			else:
				raise ValueError(f"invalid value: {value}")
		except Exception as exception:
			builtins.print(exception)
			raise


	#--------------------------------------------------------------------------------
	# 가져오기.
	#--------------------------------------------------------------------------------
	def Fetch(self) -> Optional[IRecord]:
		try:
			rawdata = self.GetRawData()
			record = rawdata.Fetch()
			if not record:
				return None
			if not record.GetFieldCount():
				return None
		except Exception as exception:
			builtins.print(exception)
			raise
			
		from .record import Record
		return Record(record)


	#--------------------------------------------------------------------------------
	# 닫기.
	#--------------------------------------------------------------------------------
	def Close(self) -> None:
		rawdata = self.GetRawData()
		rawdata.Close()